#!/usr/bin/python3

import clmctest.monitoring.LineProtocolGenerator as lp
import time
import urllib.parse
import random
import sys
import getopt
from influxdb import InfluxDBClient

# Simulation parameters
TICK_TIME = 1
DEFAULT_REQUEST_RATE_INC = 1
DEFAULT_REQUEST_RATE_INC_PERIOD = 10
SIMULATION_TIME_SEC = 60 * 60

# CLMC parameters
INFLUX_DB_URL = 'http://172.23.1.20:8086'
INFLUX_DB_NAME = 'CLMCMetrics'
AGENT1_URL = 'http://172.23.1.21:8186'
AGENT2_URL = 'http://172.23.1.22:8186'


class Sim(object):
    """

    
    Simulator for services
    """

    def __init__(self, influx_url, influx_db_name, agent1_url, agent2_url):
        """
        Sets up the simulator object

        :param influx_url: the influx DB url
        :param influx_db_name: the influx DB name
        """

        self.influx_db_name = influx_db_name
        self.agent1_url = agent1_url       
        self.agent2_url = agent2_url          

        # influx db client is created on initialisation, which will handle the influx DB queries
        url_object = urllib.parse.urlparse(influx_url)
        self.db_client = InfluxDBClient(host=url_object.hostname, port=url_object.port, database=self.influx_db_name, timeout=10)

    def reset(self):
        """
        Resets the influx db by deleting the old database and creating a new one
        """

        # Teardown DB from previous sim and bring it back up
        self.db_client.drop_database(self.influx_db_name)
        self.db_client.create_database(self.influx_db_name)

    def run(self, simulation_length_seconds):
        """
        Runs the simulation

        :param simulation_length_seconds: length of simulation
        """

        start_time = time.time() - SIMULATION_TIME_SEC
        sim_time = start_time

        # segment_size : the length of video requested at a time
        # bit_rate: MPEG-2 High 1080p 25fps = 80Mbps
        ip_endpoints = [{'agent_url': self.agent1_url, 'location': 'DC1', 'cpu': 16,
                         'mem': '8GB', 'storage': '1TB', 'request_queue': 0, 'request_arrival_rate': 0,
                         'segment_size': 2, 'video_bit_rate': 80, 'packet_size': 1500},
                        {'agent_url': self.agent2_url, 'location': 'DC2', 'cpu': 4,
                         'mem': '8GB', 'storage': '1TB', 'request_queue': 0, 'request_arrival_rate': 0,
                         'segment_size': 2, 'video_bit_rate': 80, 'packet_size': 1500}
                        ]

        # Simulate configuration of the ipendpoints
        # endpoint state->mu, sigma, secs normal distribution
        config_delay_dist = {"placing": [10, 0.68], "booting": [10, 0.68], "connecting": [10, 0.68]}

        # Simulation configuration of the media component (MC) state changes
        # "MC state", [average (sec), stddev]
        mc_config_delay_dist = { "starting": [5, 0.68], "stopping": [2, 0.68]}

        print("\nSimulation started. Generating data...")

        # Place endpoints
        max_delay = 0
        for ip_endpoint in ip_endpoints:
            agent_url = urllib.parse.urlparse(ip_endpoint["agent_url"])
            agent_db_client = InfluxDBClient(host=agent_url.hostname, port=agent_url.port, database=self.influx_db_name, timeout=10)
            delay_time = self._changeVMState(agent_db_client, sim_time, ip_endpoint, config_delay_dist['placing'][0],
                                             config_delay_dist['placing'][0] * config_delay_dist['placing'][1], 'placing')
            max_delay = max(delay_time, max_delay)
        sim_time += max_delay

        # Boot endpoints
        max_delay = 0
        for ip_endpoint in ip_endpoints:
            agent_url = urllib.parse.urlparse(ip_endpoint["agent_url"])
            agent_db_client = InfluxDBClient(host=agent_url.hostname, port=agent_url.port, database=self.influx_db_name, timeout=10)
            delay_time = self._changeVMState(agent_db_client, sim_time, ip_endpoint, config_delay_dist['booting'][0],
                                             config_delay_dist['booting'][0] * config_delay_dist['booting'][1], 'booting')
            max_delay = max(delay_time, max_delay)
        sim_time += max_delay

        # move mpegdash_service endpoints state through from 'starting' to 'running'
        max_delay = 0
        for ip_endpoint in ip_endpoints:
            agent_url       = urllib.parse.urlparse(ip_endpoint["agent_url"])
            agent_db_client = InfluxDBClient(host=agent_url.hostname, port=agent_url.port, database=self.influx_db_name, timeout=10)
            delay_avg       = mc_config_delay_dist['starting'][0]
            delay_std       = delay_avg * mc_config_delay_dist['starting'][1]
            
            delay_time = self._changeMCState(agent_db_client, sim_time, "mpegdash_service_config", delay_avg, delay_std, 0.7, 'starting', 'running')
            max_delay  = max(delay_time, max_delay)

        sim_time += max_delay

        # Connect endpoints
        max_delay = 0
        for ip_endpoint in ip_endpoints:
            agent_url = urllib.parse.urlparse(ip_endpoint["agent_url"])
            agent_db_client = InfluxDBClient(host=agent_url.hostname, port=agent_url.port, database=self.influx_db_name, timeout=10)
            delay_time = self._changeVMState(agent_db_client, sim_time, ip_endpoint, config_delay_dist['connecting'][0],
                                             config_delay_dist['connecting'][0] * config_delay_dist['connecting'][1], 'connecting')
            max_delay = max(delay_time, max_delay)
        sim_time += max_delay

        request_arrival_rate_inc = DEFAULT_REQUEST_RATE_INC
        inc_period_count = 0
        for i in range(simulation_length_seconds):
            for ip_endpoint in ip_endpoints:
                agent_url = urllib.parse.urlparse(ip_endpoint["agent_url"])
                agent_db_client = InfluxDBClient(host=agent_url.hostname, port=agent_url.port, database=self.influx_db_name, timeout=10)

                # linear inc to arrival rate
                if inc_period_count >= DEFAULT_REQUEST_RATE_INC_PERIOD:
                    ip_endpoint['request_arrival_rate'] += request_arrival_rate_inc
                    inc_period_count = 0
                else:
                    inc_period_count += 1
                # add new requests to the queue
                ip_endpoint['request_queue'] += ip_endpoint['request_arrival_rate']

                # time to process one second of video (mS) in the current second
                request_processing_time = int(random.normalvariate(10, 10 * 0.68))
                request_processing_time = max(request_processing_time, 10)
                # time depends on the length of the segments in seconds
                request_processing_time *= ip_endpoint['segment_size']

                # amount of cpu time (mS) per tick
                cpu_time_available = ip_endpoint['cpu'] * TICK_TIME * 1000
                max_requests_processed = int(cpu_time_available / request_processing_time)
                # calc how many requests processed
                if ip_endpoint['request_queue'] <= max_requests_processed:
                    # processed all of the requests
                    requests_processed = ip_endpoint['request_queue']
                else:
                    # processed the maximum number of requests
                    requests_processed = max_requests_processed

                # calculate cpu usage
                cpu_active_time = int(requests_processed * request_processing_time)
                cpu_idle_time = int(cpu_time_available - cpu_active_time)
                cpu_usage = cpu_active_time / cpu_time_available
                agent_db_client.write_points(lp.generate_cpu_report(cpu_usage, cpu_active_time, cpu_idle_time, sim_time))

                # calc network usage metrics
                bytes_rx = 2048 * requests_processed
                bytes_tx = int(
                    ip_endpoint['video_bit_rate'] / 8 * 1000000 * requests_processed * ip_endpoint['segment_size'])
                agent_db_client.write_points(lp.generate_network_report(bytes_rx, bytes_tx, sim_time))

                # time to process all of the requests in the queue
                peak_response_time = ip_endpoint['request_queue'] * request_processing_time / ip_endpoint['cpu']
                # mid-range 
                avg_response_time = (peak_response_time + request_processing_time) / 2
                agent_db_client.write_points(lp.generate_mpegdash_report('http://localhost/server-status?auto', ip_endpoint['request_arrival_rate'],
                                                                         avg_response_time, peak_response_time, sim_time))

                # need to calculate this but sent at 5mS for now
                network_request_delay = 0.005

                # calculate network response delays (2km link, 100Mbps)
                network_response_delay = self._calcNetworkDelay(2000, 100, ip_endpoint['packet_size'], ip_endpoint['video_bit_rate'])

                e2e_delay = network_request_delay + (avg_response_time / 1000) + network_response_delay
                agent_db_client.write_points(lp.generate_ipendpoint_route('http://localhost/server-status?auto', ip_endpoint['request_arrival_rate'], e2e_delay, sim_time))

                # remove requests processed off the queue
                ip_endpoint['request_queue'] -= int(requests_processed)

                # update media component state (continuously 'running')
                state_stats = {}
                state_stats['running'] = float(TICK_TIME)
                state_stats['avg_running'] = float(TICK_TIME)
                agent_db_client.write_points(lp.generate_mc_service_config("mpegdash_service_config",state_stats,sim_time))
                
            sim_time += TICK_TIME

        # Simulate tear-down of media components
        # move mpegdash_service endpoints state through from 'running' to 'stopped'
        max_delay = 0
        for ip_endpoint in ip_endpoints:
            agent_url       = urllib.parse.urlparse(ip_endpoint["agent_url"])
            agent_db_client = InfluxDBClient(host=agent_url.hostname, port=agent_url.port, database=self.influx_db_name, timeout=10)
            delay_avg       = mc_config_delay_dist['stopping'][0]
            delay_std       = delay_avg * mc_config_delay_dist['stopping'][1]
            
            delay_time = self._changeMCState(agent_db_client, sim_time, "mpegdash_service_config", delay_avg, delay_std, 0.7, 'stopping', 'stopped')
            max_delay  = max(delay_time, max_delay)

        sim_time += max_delay

        # End simulation
        end_time = sim_time
        print("Simulation Finished. Start time {0}. End time {1}. Total time {2}".format(start_time, end_time,
                                                                                         end_time - start_time))

    @staticmethod
    def _calcNetworkDelay(distance, bandwidth, packet_size, tx_video_bit_rate):
        """
        Calculates the network delay. Declared as static method since it doesn't need access to any instance variables.

        :param distance: distance metres
        :param bandwidth: bandwidth Mbps
        :param packet_size: packet size bytes
        :param tx_video_bit_rate: bp/sec
        :return: the calculated network delay
        """

        # propogation delay = distance/speed () (e.g 2000 metres * 2*10^8 for optical fibre)
        propogation_delay = distance / (2 * 100000000)
        # packetisation delay = ip packet size (bits)/tx rate (e.g. 100Mbp with  0% packet loss)
        packetisation_delay = (packet_size * 8) / (bandwidth * 1000000)
        # total number of packets to be sent
        packets = (tx_video_bit_rate * 1000000) / (packet_size * 8)

        response_delay = packets * (propogation_delay + packetisation_delay)

        return response_delay

    @staticmethod
    def _changeVMState(agent_db_client, sim_time, ip_endpoint, mu, sigma, transition_state):
        """
        Send influx data to change VM state. Declared as static method since it doesn't need access to any instance variables.

        :param sim_time: current simulation time
        :param ip_endpoint: endpoint configuration (cpu, memory, storage, etc.)
        :param mu: mean value
        :param sigma: standard deviation
        :return: the delay time
        """

        # assert transition time is at least 0
        delay_time = max(random.normalvariate(mu, sigma), 0.0)
        # transition state and state period are passed to the generate report function as keyword arguments
        agent_db_client.write_points(lp.generate_endpoint_config(ip_endpoint['cpu'], ip_endpoint['mem'], ip_endpoint['storage'], sim_time,
                                                                 **{transition_state: delay_time, 'avg_{0}'.format(transition_state): delay_time}))

        return delay_time

    @staticmethod
    def _changeMCState(agent_db_client, sim_time, mc_measurement, mu, sigma, trans_ratio, transition_state, next_state):
        """
        Send INFLUX data indicating the time taken to transition to a new state

        Returns the total time delay for the state change
        """

        # Calculate a randomized total time for the transition (and calculate relative ratios of time in transition and next state)
        total_delay_time = random.normalvariate(mu, sigma)
        transition_time = total_delay_time * trans_ratio
        next_state_time = total_delay_time - transition_time

        mc_states = {}
        
        # Report time in transition (and add the same as average)
        mc_states[transition_state] = transition_time
        mc_states["avg_" +transition_state] = transition_time

        # Report time remaining in the next state (adding the same as the average)
        mc_states[next_state] = next_state_time
        mc_states["avg_" +next_state] = next_state_time

        agent_db_client.write_points(lp.generate_mc_service_config(mc_measurement, mc_states, sim_time))

        return total_delay_time


def run_simulation(generate=True, sTime=3600):
    """
    A method which runs the data generation simulator
    :param generate: True for generating data, False for deleting the DB (optional argument, if not given, default value True is used)
    """

    global INFLUX_DB_NAME
    global INFLUX_DB_URL
    global SIMULATION_TIME_SEC
    global AGENT1_URL
    global AGENT2_URL

    SIMULATION_TIME_SEC = sTime

    simulator = Sim(INFLUX_DB_URL, INFLUX_DB_NAME, AGENT1_URL, AGENT2_URL)

    if generate:
        simulator.reset()
        simulator.run(SIMULATION_TIME_SEC)
    else:
        simulator.db_client.drop_database(simulator.influx_db_name)


if __name__ == "__main__":
    """
    The main entry for this module. Code here is executed only if the StreamingSim.py file is executed, 
    but not when it's imported in another module
    """

    # Default options for simulation (generate data and simulation time of 3600 seconds)
    genOpt = True
    simTime = 60 * 60

    # Try get some options
    try:
        opts, args = getopt.getopt( sys.argv[1:], "c:t:", ['clear','time='])

    except getopt.GetoptError:
        print( 'StreamingSim.py -c -t <seconds>' )
        sys.exit(2)
    
    # Apply options, if any
    for opt, arg in opts:
        if opt in ( '-c','--clear' ):
            genOpt = False
        
        elif opt in ('-t','--time'):
            simTime = arg

    if ( genOpt == True ):
        print( "Running simulation to generate data" )
        print( "Time period for this simulation: " + str(simTime) + " seconds" )
    else:
        print( "Clearing simulation data" )

    run_simulation( genOpt, simTime )
